// © 2020 ETH Zurich, Mechanics and Materials Lab
// © 2020 California Institute of Technology
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
// http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#pragma once

#include "ae108/cpppetsc/ParallelComputePolicy_fwd.h"
// #include <mpi.h>
#include <boost/mpi.hpp>
#include <boost/mpi/collectives.hpp>
#include <boost/serialization/array.hpp>
#include <boost/serialization/shared_ptr.hpp>
#include <boost/serialization/split_member.hpp>
#include <boost/serialization/vector.hpp>
#include <petscsys.h>

namespace ae108 {
namespace cpppetsc {

/**
 * @brief The policies configure the wrapper classes (error handling,
 * communicator).
 */
struct ParallelComputePolicy {
private:
  inline static boost::mpi::communicator _comm;
  inline static bool _initialized = false;

public:
  static void initializeCommunicator() {
    if (!_initialized) {
      MPI_Comm petscComm = PETSC_COMM_WORLD;
      _comm = boost::mpi::communicator(petscComm, boost::mpi::comm_attach);
      _initialized = true;
    }
  }

  static void finalizeCommunicator() {
    if (_initialized) {
      _comm.~communicator(); // Manually destroy the communicator
      _initialized = false;
    }
  }

  static inline MPI_Comm communicator();

  static inline void handleError(const PetscErrorCode code);

  static inline bool isPrimaryRank();

  static inline int rank();

  static inline int numberOfRanks();

  static inline void allGather(const void *sendbuf, int sendcount,
                               MPI_Datatype sendtype, void *recvbuf,
                               int recvcount, MPI_Datatype recvtype);

  static inline void barrier();

  static inline void allGatherv(const void *sendbuf, int sendcount,
                                MPI_Datatype sendtype, void *recvbuf,
                                const int *recvcounts, const int *displs,
                                MPI_Datatype recvtype);

  static inline void allReduce(const void *sendbuf, void *recvbuf, int count,
                               MPI_Datatype datatype, MPI_Op op);

  template <class EnergySamplingElement>
  static inline void shareEnergySamplingElement(
      const EnergySamplingElement &localElements,
      std::vector<EnergySamplingElement> &globalElements);

  template <class EnergySamplingElement>
  static inline void shareEnergySamplingElementVector(
      const std::vector<EnergySamplingElement> &localElements,
      std::vector<EnergySamplingElement> &globalElements);

  template <class T>
  static inline std::vector<T> all_concat_vec(const std::vector<T> &vector);
};

/********************************************************************
 *  implementations
 *******************************************************************/

#ifdef PLASTIC
MPI_Comm ParallelComputePolicy::communicator() {
  if (!_initialized) {
    initializeCommunicator(); // Ensure _comm is properly initialized
  }
  return PETSC_COMM_WORLD;
}
#else
MPI_Comm ParallelComputePolicy::communicator() { return PETSC_COMM_WORLD; }
#endif
void ParallelComputePolicy::handleError(const PetscErrorCode code) {
  CHKERRABORT(communicator(), code);
}

bool ParallelComputePolicy::isPrimaryRank() {
  int rank = 0;
  handleError(MPI_Comm_rank(communicator(), &rank));
  return rank == 0;
}

int ParallelComputePolicy::rank() {
  int rank = 0;
  handleError(MPI_Comm_rank(communicator(), &rank));
  return rank;
}

int ParallelComputePolicy::numberOfRanks() {
  int numberOfRanks = 0;
  handleError(MPI_Comm_size(communicator(), &numberOfRanks));
  return numberOfRanks;
}

void ParallelComputePolicy::allGather(const void *sendbuf, int sendcount,
                                      MPI_Datatype sendtype, void *recvbuf,
                                      int recvcount, MPI_Datatype recvtype) {
  PetscErrorCode code = MPI_Allgather(sendbuf, sendcount, sendtype, recvbuf,
                                      recvcount, recvtype, communicator());
  handleError(code);
}

void ParallelComputePolicy::barrier() {
  handleError(MPI_Barrier(communicator()));
}

void ParallelComputePolicy::allGatherv(const void *sendbuf, int sendcount,
                                       MPI_Datatype sendtype, void *recvbuf,
                                       const int *recvcounts, const int *displs,
                                       MPI_Datatype recvtype) {
  PetscErrorCode code =
      MPI_Allgatherv(sendbuf, sendcount, sendtype, recvbuf, recvcounts, displs,
                     recvtype, communicator());
  handleError(code);
}

void ParallelComputePolicy::allReduce(const void *sendbuf, void *recvbuf,
                                      int count, MPI_Datatype datatype,
                                      MPI_Op op) {
  PetscErrorCode code =
      MPI_Allreduce(sendbuf, recvbuf, count, datatype, op, communicator());
  handleError(code);
}

// Idea is to use the petsc_comm_world communicator (the original one) for
// everything and then a boost::mpi communicator for sharing the energy sampling
// elements.
template <class EnergySamplingElement>
void ParallelComputePolicy::shareEnergySamplingElement(
    const EnergySamplingElement &localElements,
    std::vector<EnergySamplingElement> &globalElements) {
  boost::mpi::all_gather(_comm, localElements, globalElements);
}

template <class EnergySamplingElement>
void ParallelComputePolicy::shareEnergySamplingElementVector(
    const std::vector<EnergySamplingElement> &localElements,
    std::vector<EnergySamplingElement> &globalElements) {

  const auto boostComm =
      boost::mpi::communicator(PETSC_COMM_WORLD, boost::mpi::comm_duplicate);
  // boost::mpi::all_reduce(boostComm, localElements, )
  barrier();
  int numProcesses = numberOfRanks();
  barrier();
  printf("rank = %u\n", _comm.rank());
  std::vector<int> localSizes(numProcesses, 0);
  int localSize = localElements.size();
  boost::mpi::all_gather(_comm, localSize, localSizes);
  std::vector<int> displacements(numProcesses, 0);
  for (int i = 1; i < numProcesses; ++i) {
    displacements[i] = displacements[i - 1] + localSizes[i - 1];
  }
  barrier();
  printf("rank = %u\n", _comm.rank());
  int totalSize = std::accumulate(localSizes.begin(), localSizes.end(), 0);
  printf("totalSize = %u\n", totalSize);
  globalElements.resize(totalSize);
  printf("localSizes = %u, %u, %u\n", localSizes[0], localSizes[1],
         localSizes[2]);
  printf("displacements = %u, %u, %u\n", displacements[0], displacements[1],
         displacements[2]);
  barrier();
  boost::mpi::all_gatherv(_comm, localElements, globalElements, localSizes,
                          displacements);
}

// template <class T> std::vector<T> ParallelComputePolicy::all_concat_vec(const
// std::vector<T> &vector)
// {
//   const auto boostComm = boost::mpi::communicator(PETSC_COMM_WORLD,
//   boost::mpi::comm_duplicate); return boost::mpi::all_reduce(boostComm,
//   vector, [](auto vec, auto other_vec) {
//     vec.insert(vec.end(), other_vec.begin(), other_vec.end());
//     return vec;
//   });
// }
template <class T>
std::vector<T>
ParallelComputePolicy::all_concat_vec(const std::vector<T> &vector) {
  const auto boostComm =
      boost::mpi::communicator(PETSC_COMM_WORLD, boost::mpi::comm_duplicate);
  auto sizes = std::vector<int>(boostComm.size());
  boost::mpi::all_gather(boostComm, static_cast<int>(vector.size()), sizes);
  const auto total_size = std::accumulate(sizes.begin(), sizes.end(), 0);
  auto concatenated = std::vector<T>(total_size);
  boost::mpi::all_gatherv(boostComm, vector, concatenated, sizes);
  return concatenated;
}
} // namespace cpppetsc
} // namespace ae108