// © 2020 ETH Zurich, Mechanics and Materials Lab
// © 2020 California Institute of Technology
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
// http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#pragma once

#include "ae108/assembly/AssemblerTypeTraits.h"
#include "ae108/cpppetsc/MeshBoundaryCondition.h"
#include "ae108/cpppetsc/TaggedVector.h"
#include "ae108/cpppetsc/clone.h"
#include "ae108/cpppetsc/LinearSolver.h"
#include "ae108/solve/dynamics/ExplicitDynamicState.h"
#include <cassert>
#include <functional>
#include <vector>

namespace ae108 {
namespace solve {

template <class Assembler, class Policy>
class ExplicitDynamicSolver {
public:
  using mesh_type = typename assembly::MeshTypeTrait<Assembler>::type;
  using vector_type = typename mesh_type::vector_type;
  using matrix_type = typename mesh_type::matrix_type;
  using state_type = dynamics::ExplicitDynamicState<cpppetsc::distributed<vector_type>>;
  using linear_solver_type = cpppetsc::LinearSolver<Policy>;

  /**
   * @param mesh A valid pointer to a mesh_type instance.
   * @param M The mass matrix.
   * @param CminusM The damping minus the mass matrix.
   * @param timestep The time step used.
   */
  explicit ExplicitDynamicSolver(const mesh_type *mesh,
                                 matrix_type M,
                                 matrix_type CminusM,
                                 const linear_solver_type *solver);

  using BoundaryConditionContainer =
      std::vector<cpppetsc::MeshBoundaryCondition<mesh_type>>;

  /**
   * @brief Calls another overload of computeSolution passing on all arguments
   * except assembler. The local force vector and the local stiffness matrix are
   * computed using assembler's assembleForceVector and assembleStiffnessMatrix
   * methods.
   *
   * @param boundaryConditions The local essential boundary conditions to apply.
   * @param state The state to start iterating at.
   * @param time This will be used to configure the assembler.
   * @param assembler Valid nonzero pointer.
   */
  state_type
  computeSolution(const BoundaryConditionContainer &boundaryConditions,
                  state_type state, const double time, 
                  const Assembler *const assembler) const; 

private:
  const mesh_type *_mesh;
  matrix_type _M;
  matrix_type _CminusM;
  const linear_solver_type *_solver;

};
} // namespace solve
} // namespace ae108

/********************************************************************
 *  implementations of methods
 *******************************************************************/

namespace ae108 {
namespace solve {
namespace detail {

template <class vector_type, class BoundaryConditionContainer>
void applyBoundaryConditions(
    const BoundaryConditionContainer &boundaryConditions,
    cpppetsc::distributed<vector_type> *const output) {

  using size_type = decltype(BoundaryConditionContainer::value_type::dof);
  using value_type = decltype(BoundaryConditionContainer::value_type::value);

  auto indices = std::vector<size_type>();
  indices.reserve(boundaryConditions.size());
  auto values = std::vector<value_type>();
  values.reserve(boundaryConditions.size());

  for (const auto &boundaryCondition : boundaryConditions) {
    const auto globalRowIndex = boundaryCondition.node.globalDofLineRange().first;
    if (globalRowIndex >= 0) {
      indices.push_back(globalRowIndex + boundaryCondition.dof);
      values.push_back(boundaryCondition.value);
    }
  }

  output->unwrap().replace().elements(indices, values);
}
} // namespace detail

template <class Assembler, class Policy>
ExplicitDynamicSolver<Assembler, Policy>::ExplicitDynamicSolver(
    const mesh_type *mesh,
    matrix_type M,
    matrix_type CminusM,
    const linear_solver_type *solver):
    _mesh(mesh), 
    _M(std::move(M)),  
    _CminusM(std::move(CminusM)),
    _solver(solver) {

  assert(_mesh);
}

template <class Assembler, class Policy>
typename ExplicitDynamicSolver<Assembler, Policy>::state_type
ExplicitDynamicSolver<Assembler, Policy>::computeSolution(
    const BoundaryConditionContainer &boundaryConditions, 
    state_type state, const double time, 
    const Assembler *const assembler) const {
  
  assert(assembler);
  
  // Assemble the global force vector
  const auto &mesh = *_mesh;
  auto localDisplacements = vector_type::fromLocalMesh(mesh);
  auto localForces = vector_type::fromLocalMesh(mesh);
  auto globalForces = vector_type::fromGlobalMesh(mesh);
  globalForces.unwrap().setZero(); // redundant
  mesh.copyToLocalVector(state.newDisplacements, &localDisplacements);
  assembler->assembleForceVector(localDisplacements, time, &localForces);
  mesh.addToGlobalVector(localForces, &globalForces);

  // Assemble the whole rhs
  auto rhs = clone(globalForces);
  rhs.unwrap().scale(-1.0);
  rhs.unwrap().addAx(_M, state.newDisplacements);
  rhs.unwrap().addAx(_CminusM, state.oldDisplacements);

  // Setup the linear solver
  auto newDisplacements = _solver->solve(rhs);

  // Apply boundary conditions
  detail::applyBoundaryConditions<vector_type>(boundaryConditions,
                                               &newDisplacements);

  // Update current and old displacements
  state_type newState = {clone(newDisplacements),
                         clone(state.newDisplacements)};

  return newState;
}

} // namespace solve
} // namespace ae108
