// © 2020 ETH Zurich, Mechanics and Materials Lab
// © 2020 California Institute of Technology
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
// http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#pragma once

#include "ae108/assembly/AssemblerTypeTraits.h"
#include "ae108/cpppetsc/MeshBoundaryCondition.h"
#include "ae108/cpppetsc/TaggedVector.h"
#include "ae108/cpppetsc/clone.h"
#include <cassert>
#include <chrono>
#include <functional>
#include <iostream>
#include <vector>

namespace ae108 {
namespace solve {

template <class Assembler> class FIRESolver {
public:
  using mesh_type = typename assembly::MeshTypeTrait<Assembler>::type;
  using policy_type = typename assembly::PolicyTypeTrait<Assembler>::type;
  using size_type = typename mesh_type::size_type;
  using value_type = typename mesh_type::value_type;
  using vector_type = typename mesh_type::vector_type;

  /**
   * @param mesh A valid nonzero pointer to a mesh_type instance.
   * @param minusM Diagonal of negative of lumped mass matrix.
   */
  explicit FIRESolver(const mesh_type *mesh, const vector_type minusM,
                      const double tolerance, const size_t maxIterations,
                      const double hMax, const size_t resetParameter = 750,
                      const bool verbose = true);

  using BoundaryConditionContainer =
      std::vector<cpppetsc::MeshBoundaryCondition<mesh_type>>;

  void changeParameters(const double tolerance, const size_t maxIterations);

  /**
   * @brief Calls computeSolution with the default assembler calls (ie. calls
   * assembleForceVector or assembleStiffnessMatrix). Passes on the rest of the
   * arguments.
   *
   * @param boundaryConditions The local essential boundary conditions to apply.
   * @param initialGuess The global vector to start iterating at.
   * @param time This will be used to configure the assembler.
   * @para assembler Valid nonzero pointer.
   */
  cpppetsc::distributed<vector_type>
  computeSolution(const BoundaryConditionContainer &boundaryConditions,
                  cpppetsc::distributed<vector_type> initialGuess,
                  const double time, const Assembler *const assembler,
                  bool &status) const;

private:
  const mesh_type *_mesh;
  vector_type _minusM;
  double _tolerance;
  size_t _maxIterations;
  const double _hMax;
  const size_t _resetParameter;
  const bool _verbose;
};
} // namespace solve
} // namespace ae108

/********************************************************************
 *  implementations
 *******************************************************************/

namespace ae108 {
namespace solve {
namespace detail {

template <class vector_type, class BoundaryConditionContainer>
void applyBoundaryConditionsToForces(
    const BoundaryConditionContainer &boundaryConditions,
    cpppetsc::distributed<vector_type> *const output) {
  using size_type = decltype(BoundaryConditionContainer::value_type::dof);
  using value_type = decltype(BoundaryConditionContainer::value_type::value);

  auto indices = std::vector<size_type>();
  indices.reserve(boundaryConditions.size());
  auto values = std::vector<value_type>();
  values.reserve(boundaryConditions.size());

  for (const auto &boundaryCondition : boundaryConditions) {
    const auto globalRowIndex =
        boundaryCondition.node.globalDofLineRange().first;
    if (globalRowIndex >= 0) {
      indices.push_back(globalRowIndex + boundaryCondition.dof);
      values.push_back(0.);
    }
  }

  output->unwrap().replace().elements(indices, values);
}

template <class vector_type, class BoundaryConditionContainer>
void applyBoundaryConditionsToDisplacements(
    const BoundaryConditionContainer &boundaryConditions,
    cpppetsc::distributed<vector_type> *const output) {
  using size_type = decltype(BoundaryConditionContainer::value_type::dof);
  using value_type = decltype(BoundaryConditionContainer::value_type::value);

  auto indices = std::vector<size_type>();
  indices.reserve(boundaryConditions.size());
  auto values = std::vector<value_type>();
  values.reserve(boundaryConditions.size());

  for (const auto &boundaryCondition : boundaryConditions) {
    const auto globalRowIndex =
        boundaryCondition.node.globalDofLineRange().first;
    if (globalRowIndex >= 0) {
      indices.push_back(globalRowIndex + boundaryCondition.dof);
      values.push_back(boundaryCondition.value);
    }
  }

  output->unwrap().replace().elements(indices, values);
}
} // namespace detail

template <class Assembler>
FIRESolver<Assembler>::FIRESolver(const mesh_type *mesh, vector_type minusM,
                                  const double tolerance,
                                  const size_t maxIterations, const double hMax,
                                  const size_t resetParameter,
                                  const bool verbose)
    : _mesh(mesh), _minusM(std::move(minusM)), _tolerance(tolerance),
      _maxIterations(maxIterations), _hMax(hMax),
      _resetParameter(resetParameter), _verbose(verbose) {
  assert(_mesh);
}

template <class Assembler>
void FIRESolver<Assembler>::changeParameters(const double tolerance,
                                             const size_t maxIterations) {
  _tolerance = tolerance;
  _maxIterations = maxIterations;
}

template <class Assembler>
cpppetsc::distributed<typename FIRESolver<Assembler>::vector_type>
FIRESolver<Assembler>::computeSolution(
    const BoundaryConditionContainer &boundaryConditions,
    cpppetsc::distributed<vector_type> initialGuess, const double time,
    const Assembler *const assembler, bool &status) const {
  auto start = std::chrono::high_resolution_clock::now();
  assert(assembler);
  status =
      false; // Initialize as false, set to true only on successful convergence

  // Paper-optimized FIRE algorithm parameters (Guénolé et al.)
  const int Nmin = 5;             // Paper recommends 5 instead of 3
  const double alphaStart = 0.25; // Paper confirms 0.25 is optimal
  const double fInc = 1.1; // More conservative than 1.25 (paper recommends 1.1)
  const double fDec = 0.5; // Paper recommends 0.5 instead of 0.6
  const double fAlpha = 0.99;          // Paper recommends 0.99 instead of 0.995
  const double dtMax = _hMax;          // More conservative max timestep
  const double dtMin = _hMax * 1e-6;   // Paper suggests higher min timestep
  const int maxConsecutiveUphill = 10; // Paper recommends 10 instead of 20

  // Tracking variables for improved convergence
  double previousResidue = 1e30;
  double minResidue = 1e30;
  int divergenceCounter = 0;
  const double divergenceThreshold =
      10; // If residue increases by more than 10%
  const int maxDivergenceCount = 3;
  size_t stagnationCounter = 0;
  auto bestDisplacements = clone(initialGuess);
  double bestResidue = 1e30;

  // History tracking for oscillation detection
  double residueHistory[5] = {0.0, 0.0, 0.0, 0.0, 0.0};
  size_t historyIndex = 0;

  // Initializations
  double dt = dtMax * 0.3; // Start with conservative time step
  double alpha = alphaStart;
  double power = 0.0;
  size_t iterations = 0;
  size_t consecutiveSteps = 0;
  int consecUphills = 0;

  // Initialize vectors
  const auto &mesh = *_mesh;
  auto displacements = clone(initialGuess);
  auto velocities = vector_type::fromGlobalMesh(mesh);
  velocities.unwrap().setZero();
  auto globalForces = vector_type::fromGlobalMesh(mesh);
  auto globalForcesNew = vector_type::fromGlobalMesh(mesh);
  auto localForces = vector_type::fromLocalMesh(mesh);
  auto localDisplacements = vector_type::fromLocalMesh(mesh);
  auto backupDisplacements = clone(initialGuess);

  // Apply boundary conditions to displacements
  detail::applyBoundaryConditionsToDisplacements<vector_type>(
      boundaryConditions, &displacements);

  // Initial force calculation
  mesh.copyToLocalVector(displacements, &localDisplacements);
  assembler->assembleForceVector(localDisplacements, time, &localForces);
  mesh.addToGlobalVector(localForces, &globalForces);

  // Safety check for force magnitude
  double forceNorm = globalForces.unwrap().norm();
  if (std::isnan(forceNorm) || std::isinf(forceNorm) || forceNorm > 1e10) {
    if (policy_type::isPrimaryRank()) {
      printf("Warning: Initial force norm is problematic: %.6e, clamping\n",
             forceNorm);
    }
    if (std::isnan(forceNorm) || std::isinf(forceNorm)) {
      globalForces.unwrap().setZero();
    } else if (forceNorm > 1e10) {
      globalForces.unwrap().scale(1e10 / forceNorm);
    }
  }

  // Convert forces to accelerations
  globalForces.unwrap().pointwiseDivide(_minusM);
  detail::applyBoundaryConditionsToForces<vector_type>(boundaryConditions,
                                                       &globalForces);

  // Initial residue check
  double residue = globalForces.unwrap().norm();
  previousResidue = residue;
  minResidue = residue;
  bestResidue = residue;
  bestDisplacements = clone(displacements);

  for (int i = 0; i < 5; i++) {
    residueHistory[i] = residue;
  }

  if (policy_type::isPrimaryRank() && _verbose) {
    printf("FIRE Solver -- Start: Iter: %lu, residue: %.8e\n", iterations,
           residue);
  }

  // Check for immediate convergence
  if (residue <= _tolerance) {
    status = true;
    if (policy_type::isPrimaryRank()) {
      printf("FIRE Solver -- Already converged! Residue: %.8e\n", residue);
    }
    return displacements;
  }

  // Adaptive tolerance strategy
  double dynamicTolerance = _tolerance;
  double initialResidue = residue;

  // Main optimization loop
  while (residue > dynamicTolerance && iterations < _maxIterations) {
    // Periodic velocity reset to help escape local minima (per paper)
    if (iterations % _resetParameter == 0 && iterations > 0) {
      if (policy_type::isPrimaryRank() && _verbose) {
        printf("Performing periodic reset at iteration %lu\n", iterations);
      }
      velocities.unwrap().setZero();
      alpha = alphaStart;
      dt = std::min(dt, dtMax * 0.1);
      consecutiveSteps = 0;
    }

    // Store current state for possible backtracking
    backupDisplacements = clone(displacements);

    // Velocity Verlet first half-step
    velocities.unwrap().plusAlphaX(0.5 * dt, globalForces);

    // Update positions
    displacements.unwrap().plusAlphaX(dt, velocities);

    // Apply boundary conditions
    detail::applyBoundaryConditionsToDisplacements<vector_type>(
        boundaryConditions, &displacements);

    // Calculate new forces
    mesh.copyToLocalVector(displacements, &localDisplacements);
    localForces.unwrap().setZero();
    assembler->assembleForceVector(localDisplacements, time, &localForces);
    globalForcesNew.unwrap().setZero();
    mesh.addToGlobalVector(localForces, &globalForcesNew);

    // Force safety check
    forceNorm = globalForcesNew.unwrap().norm();
    if (std::isnan(forceNorm) || std::isinf(forceNorm) || forceNorm > 1e10) {
      if (policy_type::isPrimaryRank()) {
        printf("Warning: Force norm is problematic: %.6e, backtracking\n",
               forceNorm);
      }
      // Backtrack to previous position
      displacements = clone(backupDisplacements);
      dt *= 0.1;
      velocities.unwrap().setZero();
      alpha = alphaStart;
      consecutiveSteps = 0;

      if (dt < dtMin) {
        if (policy_type::isPrimaryRank()) {
          printf("Error: Time step became too small (%.6e). Terminating.\n",
                 dt);
        }
        break;
      }
      continue;
    }

    // Convert forces to accelerations
    globalForcesNew.unwrap().pointwiseDivide(_minusM);
    detail::applyBoundaryConditionsToForces<vector_type>(boundaryConditions,
                                                         &globalForcesNew);

    // Velocity Verlet second half-step
    velocities.unwrap().plusAlphaX(0.5 * dt, globalForcesNew);

    // Calculate power = F·v
    power = globalForcesNew.unwrap().dotProduct(velocities);

    // Update residue for convergence check
    residue = globalForcesNew.unwrap().norm();

    // Update residue history for trend analysis
    residueHistory[historyIndex] = residue;
    historyIndex = (historyIndex + 1) % 5;

    // Track best solution
    if (residue < bestResidue) {
      bestResidue = residue;
      bestDisplacements = clone(displacements);
      stagnationCounter = 0;
    } else {
      stagnationCounter++;
    }

    // Power thresholding for small values (paper-recommended improvement)
    double powerThreshold = 1e-10 * residue;
    if (std::fabs(power) < powerThreshold) {
      if (policy_type::isPrimaryRank() && _verbose && iterations % 100 == 0) {
        printf("Power effectively zero (%.2e < %.2e), treating as uphill\n",
               power, powerThreshold);
      }
      power = -1.0; // Force uphill treatment
    }

    // Monitor residue trends
    if (iterations > 0) {
      if (residue > minResidue * divergenceThreshold && consecutiveSteps > 20) {
        // Residue has significantly increased from the minimum seen
        divergenceCounter++;

        if (policy_type::isPrimaryRank() && _verbose) {
          printf("Warning: Residue increasing: %.8e > %.8e * %.2f\n", residue,
                 minResidue, divergenceThreshold);
        }

        if (divergenceCounter >= maxDivergenceCount) {
          // Force a more aggressive reset
          if (policy_type::isPrimaryRank() && _verbose) {
            printf("Detected divergence: Performing aggressive reset\n");
          }
          displacements = clone(backupDisplacements);
          dt *= 0.2; // More aggressive timestep reduction
          velocities.unwrap().setZero();
          alpha = alphaStart;
          consecutiveSteps = 0;
          divergenceCounter = 0;

          // Break out of this iteration and try again
          continue;
        }
      } else {
        divergenceCounter = 0; // Reset counter if we're not diverging
      }
    }

    // Update minimum residue seen
    minResidue = std::min(minResidue, residue);

    if (power > 0) { // Moving downhill
      consecUphills = 0;

      // FIRE velocity update
      double fnorm = globalForcesNew.unwrap().norm();
      double vnorm = velocities.unwrap().norm();

      if (fnorm > 0.0 && vnorm > 0.0) {
        // Mix velocity with force direction
        velocities.unwrap().timesAlphaPlusBetaX(
            1.0 - alpha, alpha * vnorm / fnorm, globalForcesNew);
      }

      // Increase consecutive steps count
      consecutiveSteps++;

      // If we've done enough consecutive steps, increase timestep and decrease
      // alpha
      if (consecutiveSteps > Nmin) {
        if (residue < previousResidue * 1.001) {
          double stability_factor =
              std::min(0.1, (consecutiveSteps - Nmin) * 0.005);
          dt = std::min(dt * (fInc + stability_factor), dtMax);
          alpha *= fAlpha;
        }
      }

    } else { // Moving uphill
      consecutiveSteps = 0;
      consecUphills++;

      // Improved uphill handling with line search/backtracking
      double backtrackFactor = 0.5;
      auto tempDisplacements = clone(backupDisplacements);

      // Try a half-step
      tempDisplacements.unwrap().plusAlphaX(backtrackFactor * dt, velocities);
      detail::applyBoundaryConditionsToDisplacements<vector_type>(
          boundaryConditions, &tempDisplacements);

      // Calculate forces at half-step position
      mesh.copyToLocalVector(tempDisplacements, &localDisplacements);
      localForces.unwrap().setZero();
      assembler->assembleForceVector(localDisplacements, time, &localForces);

      auto tempForces = vector_type::fromGlobalMesh(mesh);
      tempForces.unwrap().setZero();
      mesh.addToGlobalVector(localForces, &tempForces);

      // Convert to accelerations
      tempForces.unwrap().pointwiseDivide(_minusM);
      detail::applyBoundaryConditionsToForces<vector_type>(boundaryConditions,
                                                           &tempForces);

      // Check if this smaller step is better
      double tempResidue = tempForces.unwrap().norm();

      if (tempResidue < residue) {
        // Accept the half-step if it's better
        if (policy_type::isPrimaryRank() && _verbose) {
          printf("Uphill motion: Backtracking to half-step improved residue "
                 "from %.8e to %.8e\n",
                 residue, tempResidue);
        }

        displacements = clone(tempDisplacements);
        globalForcesNew = clone(tempForces);
        residue = tempResidue;
        dt *= 0.7; // Less aggressive reduction since we found a better position
      } else {
        // Revert to original position and standard approach
        displacements = clone(backupDisplacements);
        dt *= fDec;

        // Reset velocity
        velocities.unwrap().setZero();
      }

      alpha = alphaStart;

      if (consecUphills > maxConsecutiveUphill) {
        // Too many consecutive uphill moves
        if (policy_type::isPrimaryRank() && _verbose) {
          printf("Too many uphill moves (%d), doing partial restart\n",
                 consecUphills);
        }
        dt *= 0.1;
        consecUphills = 0;

        // If things are really stuck, try more drastic measures
        if (dt < dtMin) {
          if (policy_type::isPrimaryRank()) {
            printf("Error: Time step became too small (%.6e). Terminating.\n",
                   dt);
          }
          break;
        }
      }
    }

    // Check for oscillation patterns
    if (iterations > 100 && iterations % 100 == 0) {
      bool isOscillating = false;
      bool isIncreasing = true;
      double minRes = residueHistory[0];
      double maxRes = residueHistory[0];

      for (int i = 0; i < 5; i++) {
        minRes = std::min(minRes, residueHistory[i]);
        maxRes = std::max(maxRes, residueHistory[i]);
        if (i > 0 && residueHistory[i] < residueHistory[i - 1]) {
          isIncreasing = false;
        }
      }

      isOscillating = (maxRes - minRes) > 0.1 * minRes;

      if (isOscillating) {
        dt *= 0.8;
        // velocities.unwrap().scale(0.7);
        if (policy_type::isPrimaryRank() && _verbose) {
          printf("Detected oscillation, reducing dt to %.6e at iteration %lu\n",
                 dt, iterations);
        }
      }
    }

    // Check for stagnation and try recovery
    if (stagnationCounter > 1000) {
      if (policy_type::isPrimaryRank() && _verbose) {
        printf("No improvement for 1000 iterations, trying recovery...\n");
      }

      // Try different recovery strategies based on stagnation duration
      if (stagnationCounter <= 2000) {
        // First try reduced timestep
        dt *= 0.5;
        velocities.unwrap().setZero();
        alpha = alphaStart;
      } else {
        // If still stuck, go back to best solution
        displacements = clone(bestDisplacements);
        dt = dtMax * 0.05;
        alpha = alphaStart;
        velocities.unwrap().setZero();
        stagnationCounter = 0;
      }
    }

    if (iterations % 1000 == 0 && iterations > 0) {
      // More aggressive adaptive tolerance
      if (residue < 0.02 * initialResidue && residue > _tolerance &&
          stagnationCounter > 300) { // Relaxed conditions
        dynamicTolerance = std::min(dynamicTolerance * 2.0,
                                    residue * 0.95); // More aggressive increase
        if (policy_type::isPrimaryRank() && _verbose) {
          printf("Adjusting dynamic tolerance to %.8e\n", dynamicTolerance);
        }
      }
    }

    // Copy new forces to current forces for next iteration
    globalForces = clone(globalForcesNew);

    // Store current residue for next iteration
    previousResidue = residue;

    // Increment iteration counter
    iterations++;

    // Output progress
    if (_verbose && (iterations % 250 == 0) && policy_type::isPrimaryRank()) {
      printf("FIRE Solver -- Iter: %lu, residue: %.8e, power: %.8e, dt: %.8e, "
             "alpha: %.8e\n",
             iterations, residue, power, dt, alpha);
    }
  }

  // Consider using best solution found if current one is worse
  if (bestResidue < residue * 0.9) {
    if (policy_type::isPrimaryRank() && _verbose) {
      printf("Using best solution found (%.8e vs %.8e)\n", bestResidue,
             residue);
    }
    displacements = clone(bestDisplacements);
    residue = bestResidue;
  }

  // Set status based on convergence
  status = (residue <= _tolerance || residue <= dynamicTolerance);

  // Record timing information
  auto end = std::chrono::high_resolution_clock::now();
  std::chrono::duration<double> duration = end - start;

  if (policy_type::isPrimaryRank()) {
    printf("FIRE Solver -- Final: Iter: %lu, residue: %.8e\n", iterations,
           residue);
    printf("Solver took: %.6f seconds\n", duration.count());

    if (!status) {
      printf("Warning: FIRE solver did not converge within %lu iterations\n",
             _maxIterations);
    }
  }

  return displacements;
}

} // namespace solve
} // namespace ae108
