// © 2020 ETH Zurich, Mechanics and Materials Lab
// © 2020 California Institute of Technology
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
// http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#pragma once

#include "ae108/assembly/AssemblerTypeTraits.h"
#include "ae108/cpppetsc/MeshBoundaryCondition.h"
#include "ae108/cpppetsc/TaggedVector.h"
#include "ae108/cpppetsc/clone.h"
#include "ae108/solve/AffineTransform.h"
#include "ae108/cpppetsc/LinearSolver.h"
#include "ae108/solve/dynamics/ExplicitDynamicState.h"
#include <cassert>
#include <functional>
#include <vector>

namespace ae108 {
namespace solve {

template <class Assembler, class Policy>
class TransformingExplicitDynamicSolver {
public:
  using mesh_type = typename assembly::MeshTypeTrait<Assembler>::type;
  using vector_type = typename mesh_type::vector_type;
  using matrix_type = typename mesh_type::matrix_type;
  using state_type = dynamics::ExplicitDynamicState<cpppetsc::distributed<vector_type>>;
  using linear_solver_type = cpppetsc::LinearSolver<Policy>;

  /**
   * @param mesh A valid pointer to a mesh_type instance.
   * @param M The mass matrix.
   * @param MPlusC The damping plus the mass matrix.
   * @param CMinusM The damping minus the mass matrix.
   */
  explicit TransformingExplicitDynamicSolver(const mesh_type *mesh,
                                             matrix_type M,
                                             matrix_type MPlusC,
                                             matrix_type CMinusM);

  using BoundaryConditionContainer = AffineTransform<Policy>;

  /**
   * @brief Calls another overload of computeSolution passing on all arguments
   * except assembler. The local force vector and the local stiffness matrix are
   * computed using assembler's assembleForceVector and assembleStiffnessMatrix
   * methods.
   *
   * @param transform The affine transform to apply.
   * @param state The state to start iterating at.
   * @param time This will be used to configure the assembler.
   * @param assembler Valid nonzero pointer.
   */
  state_type
  computeSolution(const BoundaryConditionContainer &transform,
                  state_type state, const double time, 
                  const Assembler *const assembler) const; 

private:
  const mesh_type *_mesh;
  matrix_type _M;
  matrix_type _MPlusC;
  matrix_type _CMinusM;

};
} // namespace solve
} // namespace ae108

/********************************************************************
 *  implementations of methods
 *******************************************************************/

#include "ae108/cpppetsc/createTransformInput.h"

namespace ae108 {
namespace solve {

template <class Assembler, class Policy>
TransformingExplicitDynamicSolver<Assembler, Policy>::TransformingExplicitDynamicSolver(
    const mesh_type *mesh,
    matrix_type M,
    matrix_type MPlusC,
    matrix_type CMinusM):
    _mesh(mesh), 
    _M(std::move(M)),  
    _MPlusC(std::move(MPlusC)),
    _CMinusM(std::move(CMinusM)) {

  assert(_mesh);
}

template <class Assembler, class Policy>
typename TransformingExplicitDynamicSolver<Assembler, Policy>::state_type
TransformingExplicitDynamicSolver<Assembler, Policy>::computeSolution(
    const BoundaryConditionContainer &transform,
    state_type state, const double time, 
    const Assembler *const assembler) const {
  
  assert(assembler);
  
  // Compute the full new displacements from the transformed (reduced) ones
  auto fullNewDisplacements = clone(transform.vector);
  apply(transform, state.newDisplacements, &fullNewDisplacements);

  // Assemble the global force vector
  const auto &mesh = *_mesh;
  auto localDisplacements = vector_type::fromLocalMesh(mesh);
  auto localForces = vector_type::fromLocalMesh(mesh);
  auto fullForces = vector_type::fromGlobalMesh(mesh);
  fullForces.unwrap().setZero();
  mesh.copyToLocalVector(fullNewDisplacements, &localDisplacements);
  assembler->assembleForceVector(localDisplacements, time, &localForces);
  mesh.addToGlobalVector(localForces, &fullForces);
  auto reducedForces = createTransformInput(transform.matrix);
  Policy::handleError(MatMultTranspose(transform.matrix.data(),
                                       fullForces.unwrap().data(),
                                       reducedForces.unwrap().data()));

  // Assemble transformed matrices
  auto reducedM = matrix_type::fromPtAP(transform.matrix, _M);
  auto reducedMPlusC = matrix_type::fromPtAP(transform.matrix, _MPlusC);
  auto reducedCMinusM = matrix_type::fromPtAP(transform.matrix, _CMinusM);

  // Assemble the whole rhs
  auto rhs = clone(reducedForces);
  rhs.unwrap().scale(-1.0);
  rhs.unwrap().addAx(reducedM, state.newDisplacements);
  rhs.unwrap().addAx(reducedCMinusM, state.oldDisplacements);

  // Setup the linear solver
  const auto linearSolver = linear_solver_type(&reducedMPlusC);

  // Solve for reduced displacements
  auto newDisplacements = linearSolver.solve(rhs);

  // Update current and old displacements
  state_type newState = {clone(newDisplacements),
                         clone(state.newDisplacements)};
  return newState;
}

} // namespace solve
} // namespace ae108
