// © 2020 ETH Zurich, Mechanics and Materials Lab
// © 2020 California Institute of Technology
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
// http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#pragma once

#include "ae108/assembly/AssemblerTypeTraits.h"
#include "ae108/cpppetsc/MeshBoundaryCondition.h"
#include "ae108/cpppetsc/TaggedVector.h"
#include "ae108/solve/AffineTransform.h"
#include <cassert>
#include <functional>
#include <vector>

namespace ae108 {
namespace solve {

template <class Assembler> class TransformingFIRESolver {
public:
  using policy_type = typename assembly::PolicyTypeTrait<Assembler>::type;
  using mesh_type = typename assembly::MeshTypeTrait<Assembler>::type;
  using vector_type = typename mesh_type::vector_type;
  using matrix_type = typename mesh_type::matrix_type;

  /**
   * @param mesh A valid pointer to a mesh_type instance.
   * @param minusM The negative of the mass matrix.
   * @param tolerance The tolerance for the solver.
   * @param maxIterations The maximum iterations before giving up.
   * @param hMax The maximum inverse pseudo time step.
   */
  explicit TransformingFIRESolver(const mesh_type *mesh,
                                  const vector_type minusM,
                                  const double tolerance = 1e-8,
                                  const size_t maxIterations = 1e5,
                                  const double hMax = 1e-3);

  using BoundaryConditionContainer = AffineTransform<policy_type>;

  /**
   * @brief Calls another overload of computeSolution passing on all arguments
   * except assembler. The local force vector and the local stiffness matrix are
   * computed using assembler's assembleForceVector and assembleStiffnessMatrix
   * methods.
   *
   * @param transform The affine transform to apply.
   * @param initialGuess The state to start iterating at.
   * @param time This will be used to configure the assembler.
   * @param assembler Valid nonzero pointer.
   */
  cpppetsc::distributed<vector_type>
  computeSolution(const BoundaryConditionContainer &transform,
                  cpppetsc::distributed<vector_type> initialGuess,
                  const double time, const Assembler *const assembler,
                  bool & status) const; 

private:
  const mesh_type *_mesh;
  vector_type _minusM;
  const double _tolerance;
  const size_t _maxIterations;
  const double _hMax;

};
} // namespace solve
} // namespace ae108

/********************************************************************
 *  implementations of methods
 *******************************************************************/

#include "ae108/cpppetsc/createTransformInput.h"

namespace ae108 {
namespace solve {

template <class Assembler>
TransformingFIRESolver<Assembler>::TransformingFIRESolver(
    const mesh_type *mesh,
    vector_type minusM,
    const double tolerance,
    const size_t maxIterations,
    const double hMax):
    _mesh(mesh), 
    _minusM(std::move(minusM)), 
    _tolerance(tolerance), 
    _maxIterations(maxIterations),
    _hMax(hMax) {

  assert(_mesh);
}

template <class Assembler>
cpppetsc::distributed<typename TransformingFIRESolver<Assembler>::vector_type>
TransformingFIRESolver<Assembler>::computeSolution(
    const BoundaryConditionContainer &transform,
    cpppetsc::distributed<vector_type> initialGuess, 
    const double time, const Assembler *const assembler,
    bool & status) const {
  
  assert(assembler);
  
    // Algorithm parameters
  const int Nmin = 5;  
  const double alphaStart = 0.1; 
  const double fInc = 1.1; 
  const double fDec = 0.5; 
  const double fAlpha = 0.99; 
  
  // Initializations
  double h = _hMax;
  double alpha = alphaStart;
  double power = 0.0;
  size_t iterations = 0;	
  size_t N = 0;

  // Initialize global and local quantities
  const auto &mesh = *_mesh;
  auto reducedDisplacements = clone(initialGuess);
  auto reducedVelocities = createTransformInput(transform.matrix);
  auto reducedForces = createTransformInput(transform.matrix);
  auto reducedForcesNew = createTransformInput(transform.matrix);
  auto globalForces = vector_type::fromGlobalMesh(mesh);
  auto localForces = vector_type::fromLocalMesh(mesh);
  auto localDisplacements = vector_type::fromLocalMesh(mesh);

  // First step of velocity Verlet ----------------------------------------------------
  // Compute forces 
  auto fullDisplacements = clone(transform.vector);
  apply(transform, reducedDisplacements, &fullDisplacements);
  mesh.copyToLocalVector(fullDisplacements, &localDisplacements);
  localForces.unwrap().setZero();
  assembler->assembleForceVector(localDisplacements, time, &localForces);
  globalForces.unwrap().setZero();
  mesh.addToGlobalVector(localForces, &globalForces);
  globalForces.unwrap().pointwiseDivide(_minusM); // convert to accelerations
  // Need to zero out reducedForces?
  policy_type::handleError(MatMultTranspose(transform.matrix.data(),
                                       globalForces.unwrap().data(),
                                       reducedForces.unwrap().data()));

  // Compute new displacements
  reducedDisplacements.unwrap().timesAlphaPlusBetaX(1, 0.5*h*h, reducedForces);

  // Update iterations
  N++; iterations++;
  
  // Compute new forces 
  apply(transform, reducedDisplacements, &fullDisplacements);
  mesh.copyToLocalVector(fullDisplacements, &localDisplacements);
  localForces.unwrap().setZero();
  assembler->assembleForceVector(localDisplacements, time, &localForces);
  globalForces.unwrap().setZero();
  mesh.addToGlobalVector(localForces, &globalForces);
  globalForces.unwrap().pointwiseDivide(_minusM); // convert to accelerations
  // Need to zero out reducedForcesNew?
  policy_type::handleError(MatMultTranspose(transform.matrix.data(),
                                       globalForces.unwrap().data(),
                                       reducedForcesNew.unwrap().data()));

  // Compute new velocities
  reducedVelocities.unwrap().timesAlphaPlusBetaXPlusGammaY(1, 
    0.5*h, reducedForces, 0.5*h, reducedForcesNew);

  // Calculate residue
  double residue = reducedForcesNew.unwrap().norm();

  // Keep track of uphills
  size_t N_uphills = 0;

  // Start iterations ----------------------------------------------------------------------
  while(residue > _tolerance && iterations < _maxIterations){

    // Swap forces
    reducedForces = clone(reducedForcesNew);
    
    // Calculate power
    power = reducedForces.unwrap().dotProduct(reducedVelocities);

    // Velocity mixing
    reducedVelocities.unwrap().timesAlphaPlusBetaX(1-alpha, 
      alpha*reducedVelocities.unwrap().norm()/reducedForces.unwrap().norm(), reducedForces);
   
    if (power > 0 && N > Nmin){
      h = std::min(h*fInc, _hMax);
      alpha = alpha*fAlpha;
    } else if (power <= 0){
      // printf("----- Uphill!-----\n");
      N_uphills++;
      h = h*fDec;
      // Correct uphill motion
      reducedDisplacements.unwrap().plusAlphaX(-0.5*h, reducedVelocities);   
      reducedVelocities.unwrap().setZero();
      alpha = alphaStart;
      N = 0;
    }
    
    // Compute new displacements
    reducedDisplacements.unwrap().timesAlphaPlusBetaXPlusGammaY(1, 
      h, reducedVelocities, 0.5*h*h, reducedForces);

    // Update iterations
    N++; iterations++;

    // Compute new forces
    apply(transform, reducedDisplacements, &fullDisplacements);
    mesh.copyToLocalVector(fullDisplacements, &localDisplacements);
    localForces.unwrap().setZero();
    assembler->assembleForceVector(localDisplacements, time, &localForces);
    globalForces.unwrap().setZero();
    mesh.addToGlobalVector(localForces, &globalForces);
    globalForces.unwrap().pointwiseDivide(_minusM); // convert to accelerations
    // Need to zero out reducedForcesNew?
    policy_type::handleError(MatMultTranspose(transform.matrix.data(),
                                         globalForces.unwrap().data(),
                                         reducedForcesNew.unwrap().data()));
       
    // Compute new velocities
    reducedVelocities.unwrap().timesAlphaPlusBetaXPlusGammaY(1, 
      0.5*h, reducedForces, 0.5*h, reducedForcesNew);

    // Check convergence        
    residue = reducedForcesNew.unwrap().norm();
  
    if (iterations%1000 == 0) {
      printf("Iter: %lu - Residual = %.8f\n", iterations, residue);
    }
    
    if(std::isnan(residue)){
      printf("Error: FIRE solver residue is NaN\n"); exit(1);
    }
  }

  // if (iterations == _maxIterations) {
  //   printf("Error: FIRE solver could not converge in %lu " "iterations.\n", _maxIterations);exit(1);
  // }
  printf("Residual = %.8f at %lu iterations\n", residue, iterations);
  printf("N_uphills = %lu\n", N_uphills);

  status = (residue <= _tolerance) ? true : false;

  return reducedDisplacements;
}

} // namespace solve
} // namespace ae108
