// © 2020 ETH Zurich, Mechanics and Materials Lab
// © 2020 California Institute of Technology
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
// http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#pragma once

#include "ae108/assembly/AssemblerTypeTraits.h"
#include "ae108/cpppetsc/MeshBoundaryCondition.h"
#include "ae108/cpppetsc/TAOSolverWithStatus.h"
#include "ae108/cpppetsc/TaggedVector.h"
#include "ae108/cpppetsc/clone.h"
#include "ae108/solve/AffineTransform.h"
#include <cassert>
#include <functional>
#include <vector>

namespace ae108 {
namespace solve {

template <class Assembler> class TransformingTaoSolver {
public:
  using policy_type = typename assembly::PolicyTypeTrait<Assembler>::type;
  using mesh_type = typename assembly::MeshTypeTrait<Assembler>::type;
  using size_type = typename mesh_type::size_type;
  using value_type = typename mesh_type::value_type;
  using real_type = typename mesh_type::real_type;
  using vector_type = typename mesh_type::vector_type;
  using matrix_type = typename mesh_type::matrix_type;

  /**
   * @param mesh A valid nonzero pointer to a mesh_type instance.
   */
  explicit TransformingTaoSolver(const mesh_type *mesh);

  using BoundaryConditionContainer = AffineTransform<policy_type>;

  /**
   * @brief Calls computeSolution with the default assembler calls (ie. calls
   * assembleForceVector or assembleStiffnessMatrix). Passes on the rest of the
   * arguments.
   *
   * @param transform The affine transform to apply.
   * @param initialGuess The global (untransformed) vector to start iterating
   * at.
   * @param time This will be used to configure the assembler.
   * @para assembler Valid nonzero pointer.
   */
  cpppetsc::distributed<vector_type>
  computeSolution(const BoundaryConditionContainer &transform,
                  cpppetsc::distributed<vector_type> initialGuess,
                  const double time, const Assembler *const assembler,
                  const double bound, bool & status) const;

  using LocalEnergyAssembler = std::function<void(
      const cpppetsc::local<vector_type> &, double, real_type *)>;

  using LocalForceVectorAssembler =
      std::function<void(const cpppetsc::local<vector_type> &, double,
                         cpppetsc::local<vector_type> *)>;

  using LocalStiffnessMatrixAssembler = std::function<void(
      const cpppetsc::local<vector_type> &, double, matrix_type *)>;

  /**
   * @brief Calls computeSolution wrapping up the local assembler calls into
   * distributed assembler calls. Passes on the rest of the arguments.
   *
   * @param transform The affine transform to apply.
   * @param initialGuess The global (untransformed) vector to start iterating
   * at.
   * @param time This will be used to configure the assembler.
   * @param assembleEnergy A valid callable. It will be called to assemble
   * the local energy.
   * @param assembleForceVector A valid callable. It will be called to assemble
   * the local force vector.
   * @param assembleStiffnessMatrix A valid callable. It will be called to
   * assemble the stiffness matrix.
   */
  cpppetsc::distributed<vector_type>
  computeSolution(const BoundaryConditionContainer &transform,
                  cpppetsc::distributed<vector_type> initialGuess,
                  const double time, LocalEnergyAssembler assembleEnergy,
                  LocalForceVectorAssembler assembleForceVector,
                  LocalStiffnessMatrixAssembler assemblerStiffnessMatrix,
                  const double bound, bool & status) const;

  using DistributedEnergyAssembler = std::function<void(
      const cpppetsc::distributed<vector_type> &, double, real_type *)>;
      
  using DistributedForceVectorAssembler =
      std::function<void(const cpppetsc::distributed<vector_type> &, double,
                         cpppetsc::distributed<vector_type> *)>;

  using DistributedStiffnessMatrixAssembler = std::function<void(
      const cpppetsc::distributed<vector_type> &, double, matrix_type *)>;

  /**
   * @brief The goal of this function is to find the minimizer y of
   * E(P * y + c) -> min, where (P, c) is the provided affine transformation
   * `transform`. The transform can be used to specify boundary conditions by
   * fixing degrees of freedom in the output of the transform.
   *
   * More precisely, this function solves the equation E' = 0, which is
   * equivalent to P^t * F(P * y + c) = 0, where F(P * y + c) are the forces at
   * P * y + c.
   *
   * @param transform The affine transform to apply.
   * @param initialGuess The global (untransformed) vector to start iterating
   * at.
   * @param time This will be used to configure the assembler.
   * @param assembleEnergy A valid callable. It will be called to assemble
   * the distributed energy.
   * @param assembleForceVector A valid callable. It will be called to assemble
   * the distributed force vector.
   * @param assembleStiffnessMatrix A valid callable. It will be called to
   * assemble the stiffness matrix.
   */
  cpppetsc::distributed<vector_type> computeSolution(
      const BoundaryConditionContainer &transform,
      cpppetsc::distributed<vector_type> initialGuess, const double time,
      DistributedEnergyAssembler assembleEnergy,
      DistributedForceVectorAssembler assembleForceVector,
      DistributedStiffnessMatrixAssembler assemblerStiffnessMatrix,
      const double bound, bool & status) const;

private:
  const mesh_type *_mesh;
};

} // namespace solve
} // namespace ae108

/********************************************************************
 *  implementations
 *******************************************************************/

#include "ae108/cpppetsc/createTransformInput.h"

namespace ae108 {
namespace solve {

template <class Assembler>
TransformingTaoSolver<Assembler>::TransformingTaoSolver(const mesh_type *mesh)
    : _mesh(mesh) {}

template <class Assembler>
cpppetsc::distributed<typename TransformingTaoSolver<Assembler>::vector_type>
TransformingTaoSolver<Assembler>::computeSolution(
    const BoundaryConditionContainer &transform,
    cpppetsc::distributed<vector_type> initialGuess, const double time,
    const Assembler *const assembler,
    const double bound, bool & status) const {
  assert(assembler);
  return computeSolution(
      transform, std::move(initialGuess), time,
      LocalEnergyAssembler(
          [assembler](const cpppetsc::local<vector_type> &localDisplacements,
                      const double time, real_type *const localEnergy) {
            assembler->assembleEnergy(localDisplacements, time, localEnergy);
          }),
      LocalForceVectorAssembler(
          [assembler](const cpppetsc::local<vector_type> &localDisplacements,
                      const double time,
                      cpppetsc::local<vector_type> *const localForces) {
            assembler->assembleForceVector(localDisplacements, time,
                                           localForces);
          }),
      LocalStiffnessMatrixAssembler(
          [assembler](const cpppetsc::local<vector_type> &localDisplacements,
                      const double time, matrix_type *const output) {
            assembler->assembleStiffnessMatrix(localDisplacements, time,
                                               output);
          }),
      bound, status);
}

template <class Assembler>
cpppetsc::distributed<typename TransformingTaoSolver<Assembler>::vector_type>
TransformingTaoSolver<Assembler>::computeSolution(
    const BoundaryConditionContainer &transform,
    cpppetsc::distributed<vector_type> initialGuess, const double time,
    LocalEnergyAssembler assembleEnergy,
    LocalForceVectorAssembler assembleForceVector,
    LocalStiffnessMatrixAssembler assembleStiffnessMatrix,
    const double bound, bool & status) const {
  assert(assembleEnergy);
  assert(assembleForceVector);
  assert(assembleStiffnessMatrix);

  const auto &mesh = *_mesh;

  auto localDisplacements = vector_type::fromLocalMesh(mesh);
  auto localForces = vector_type::fromLocalMesh(mesh);

  const auto distributedEnergyAssembler =
      [&assembleEnergy, &mesh,
       &localDisplacements](const cpppetsc::distributed<vector_type> &input,
                            const double time, real_type *const output) {
        mesh.copyToLocalVector(input, &localDisplacements);

        *output = real_type{};
        assembleEnergy(localDisplacements, time, output);

        policy_type::handleError(MPI_Allreduce(MPI_IN_PLACE, output, 1,
                                               MPIU_REAL, MPIU_SUM,
                                               policy_type::communicator()));
      };

  const auto distributedForceVectorAssembler =
      [&assembleForceVector, &mesh, &localDisplacements, &localForces](
          const cpppetsc::distributed<vector_type> &input, const double time,
          cpppetsc::distributed<vector_type> *const output) {
        mesh.copyToLocalVector(input, &localDisplacements);

        localForces.unwrap().setZero();
        assembleForceVector(localDisplacements, time, &localForces);

        mesh.addToGlobalVector(localForces, output);
      };

  const auto distributedStiffnessMatrixAssembler =
      [&assembleStiffnessMatrix, &mesh,
       &localDisplacements](const cpppetsc::distributed<vector_type> &input,
                            const double time, matrix_type *const output) {
        mesh.copyToLocalVector(input, &localDisplacements);

        assembleStiffnessMatrix(localDisplacements, time, output);
        output->finalize();
      };

  return computeSolution(transform, std::move(initialGuess), time,
                         distributedEnergyAssembler,
                         distributedForceVectorAssembler,
                         distributedStiffnessMatrixAssembler,
                         bound, status);
}

template <class Assembler>
cpppetsc::distributed<typename TransformingTaoSolver<Assembler>::vector_type>
TransformingTaoSolver<Assembler>::computeSolution(
    const BoundaryConditionContainer &transform,
    cpppetsc::distributed<vector_type> initialGuess, const double time,
    DistributedEnergyAssembler assembleEnergy,
    DistributedForceVectorAssembler assembleForceVector,
    DistributedStiffnessMatrixAssembler assembleStiffnessMatrix,
    const double bound, bool & status) const {
  assert(assembleEnergy);
  assert(assembleForceVector);
  assert(assembleStiffnessMatrix);

  auto fullInput = clone(transform.vector);

  auto objective = [&](const cpppetsc::distributed<vector_type> &input,
                      double *const output) {
    apply(transform, input, &fullInput);
    *output = real_type{};
    assembleEnergy(fullInput, time, output);
  };

  auto fullForces = vector_type::fromGlobalMesh(*_mesh);
  auto gradient = [&](const cpppetsc::distributed<vector_type> &input,
                      cpppetsc::distributed<vector_type> *const output) {
    apply(transform, input, &fullInput);

    fullForces.unwrap().setZero();
    assembleForceVector(fullInput, time, &fullForces);
    policy_type::handleError(MatMultTranspose(transform.matrix.data(),
                                              fullForces.unwrap().data(),
                                              output->unwrap().data()));
  };

  auto fullStiffnessMatrix = matrix_type::fromMesh(*_mesh);
  auto jacobian = [&](const cpppetsc::distributed<vector_type> &input,
                      matrix_type *const output) {
    apply(transform, input, &fullInput);

    fullStiffnessMatrix.setZero();
    assembleStiffnessMatrix(fullInput, time, &fullStiffnessMatrix);
    auto outputData = output->data();
    policy_type::handleError(MatPtAP(fullStiffnessMatrix.data(),
                                     transform.matrix.data(), MAT_REUSE_MATRIX,
                                     PETSC_DEFAULT, &outputData));
    assert(outputData == output->data() && "PETSc does not change the pointer");
  };

  using Solver = cpppetsc::TAOSolver<policy_type>;

  auto lowerBound = cpppetsc::tag<cpppetsc::DistributedTag>(
      vector_type::fromLayoutOf(initialGuess));
  lowerBound.unwrap().fill(-bound);
  lowerBound.unwrap().timesAlphaPlusBetaX(1,1,initialGuess);

  auto upperBound = cpppetsc::tag<cpppetsc::DistributedTag>(
      vector_type::fromLayoutOf(initialGuess));
  upperBound.unwrap().fill(bound);
  upperBound.unwrap().timesAlphaPlusBetaX(1,1,initialGuess);

  Solver solver{matrix_type::fromPtAP(transform.matrix, fullStiffnessMatrix)};

  solver.setBounds(lowerBound, upperBound);

  double residual = 0.;
  return solver.solve(std::move(objective), 
                      std::move(gradient),
                      std::move(jacobian),
                      std::move(initialGuess),
                      status, residual);
}

} // namespace solve
} // namespace ae108
